import sys
import stat
import pytest
import pathlib
import mygaiadb
from mygaiadb.query import LocalGaiaSQL
from mygaiadb.data import download, compile
import pandas as pd


@pytest.mark.order(1)
def test_download():
    download.download_gaia_source(test=True)
    download.download_2mass_best_neightbour(test=True)
    download.download_allwise_best_neightbour(test=True)
    download.download_gaia_astrophysical_parameters(test=True)
    download.download_2mass(test=True)
    # download.download_allwise(test=True)


@pytest.mark.order(2)
def test_compile():
    compile.compile_gaia_sql_db(indexing=False)
    compile.compile_tmass_sql_db(indexing=False)
    # compile.compile_allwise_sql_db(indexing=False)
    # check if database exist
    assert mygaiadb.gaia_sql_db_path.exists()
    assert mygaiadb.tmass_sql_db_path.exists()
    # assert mygaiadb.allwise_sql_db_path.exists()
    # assert database > 2GB
    assert mygaiadb.gaia_sql_db_path.stat().st_size > 2e+9


@pytest.mark.order(3)
def test_user_table():
    localdb = LocalGaiaSQL(load_allwise=False)
    localdb.upload_user_table(pd.DataFrame({"source_id": [5188146770731873152, 4611686018427432192, 5764607527332179584]}), "user_table_1")
    result = localdb.query("""SELECT * FROM user_table.user_table_1""")
    assert result["source_id"].tolist() == [5188146770731873152, 4611686018427432192, 5764607527332179584]
    localdb.remove_user_table("user_table_1")
    assert localdb.list_user_tables() == {}

@pytest.mark.order(4)
def test_utilities():
    localdb = LocalGaiaSQL(load_allwise=False)
    localdb.list_all_tables()
    localdb.get_table_column("gaiadr3.gaia_source")
    localdb.get_table_column("gaiadr3.tmasspscxsc_best_neighbour")
    localdb.get_table_column("tmass.twomass_psc")


@pytest.mark.order(5)
def test_query():
    query = """
    SELECT * 
    FROM gaiadr3.gaia_source as G
    INNER JOIN gaiadr3.astrophysical_parameters as GA on GA.source_id = G.source_id
    INNER JOIN gaiadr3.tmasspscxsc_best_neighbour as T on G.source_id = T.source_id
    INNER JOIN gaiadr3.allwise_best_neighbour as W on W.source_id = T.source_id  
    INNER JOIN tmass.twomass_psc as TM on TM.designation = T.original_ext_source_id
    WHERE (G.has_xp_continuous = 1)  
    LIMIT 10
    """
    localdb = LocalGaiaSQL(load_allwise=False)
    localdb.save_csv(query, "output.csv")
    query_df = localdb.query(query)



import inspect
import warnings
import numpy as np
from ..utils import radec_to_ecl

try:
    from zero_point import zpt

    _have_zpt = True
except ImportError:
    _have_zpt = False

try:
    import mwdust

    _have_mwdust = True
except ImportError:
    _have_mwdust = False

try:
    from galpy.orbit import Orbit
    from galpy.util.coords import radec_to_lb

    _have_galpy = True
except ImportError:
    _have_galpy = False


class QueryCallback:
    """
    Callback to add new column to SQL query on the fly
    """

    def __init__(self, new_col_name, func):
        """
        INPUT:
            new_col_name (string): Name of the new column you wan to add
            func (function): function that maps query columns to new columns, arguements of this function need to have \
                the same names as columns in query
        """
        self.new_col_name = new_col_name
        self.func = func
        self.required_col = list(inspect.getfullargspec(self.func))[0]


class ZeroPointCallback(QueryCallback):
    def __init__(self, new_col_name="parallax_w_zp"):
        """
        Callback to use ``gaiadr3_zeropoint`` to get zero-point corrected parallax

        INPUT:
            new_col_name (string): Name of the new column you wan to add
        """
        super().__init__(new_col_name, self.parallax_zp_func)
        if not _have_zpt:
            raise RuntimeError(
                "You need to have gaiadr3_zeropoint package installed to use this callback. Please see: https://gitlab.com/icc-ub/public/gaiadr3_zeropoint"
            )
        zpt.load_tables()

    @staticmethod
    def parallax_zp_func(
        ra,
        dec,
        parallax,
        phot_bp_mean_mag,
        nu_eff_used_in_astrometry,
        pseudocolour,
        astrometric_params_solved,
    ):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", UserWarning)
            ect_lon, ect_lat = radec_to_ecl(ra, dec)
            return parallax - zpt.get_zpt(
                phot_bp_mean_mag,
                nu_eff_used_in_astrometry,
                pseudocolour,
                ect_lat,
                astrometric_params_solved,
            )


class DustCallback(QueryCallback):
    def __init__(self, new_col_name="sfd_ebv", filter=None, dustmap="SFD"):
        """
        Callback to use ``mwdust`` to get extinction

        INPUT:
            new_col_name (string): Name of the new column you wan to add
            filter (string): extinction in which filter, see mwdust
            dustmap (string): which dust map to use, currently only supporting "SFD"
        """
        if not _have_mwdust:
            raise RuntimeError(
                "You need to have mwdust package installed to use this callback. Please see: https://github.com/jobovy/mwdust"
            )
        if not _have_galpy:
            raise RuntimeError(
                "You need to have galpy package installed to use this callback. Please see: https://github.com/jobovy/galpy"
            )
        self.filter = filter
        if dustmap.lower() == "sfd":
            self.sfd = mwdust.SFD(filter=self.filter, noloop=True)
            self._func_wrapped = lambda ra, dec: self.sfd_ebv_func(ra, dec)
        super().__init__(new_col_name, self._func_wrapped)

    def sfd_ebv_func(self, ra, dec):
        lb = radec_to_lb(ra, dec, degree=True)
        l, b = lb[:, 0], lb[:, 1]
        return self.sfd(l, b, np.ones_like(l))


class OrbitsCallback(QueryCallback):
    def __init__(self, new_col_name="e"):
        """
        Callback to use ``galpy`` to setup orbit

        INPUT:
            new_col_name (string): Name of the new column you wan to add
            filter (string): extinction in which filter, see mwdust
            dustmap (string): which dust map to use, currently only supporting "SFD"
        """
        if not _have_galpy:
            raise RuntimeError(
                "You need to have galpy package installed to use this callback. Please see: https://github.com/jobovy/galpy"
            )
        _r0 = 8.23
        _v0 = 249.44
        _z0 = 0.0208

        self._func_wrapped = (
            lambda ra, dec, pmra, pmdec, parallax, radial_velocity: Orbit(
                [
                    ra,
                    dec,
                    (1 / parallax),
                    pmra,
                    pmdec,
                    radial_velocity,
                ],
                radec=True,
                ro=_r0,
                vo=_v0,
                zo=_z0,
            )
        )
        super().__init__(new_col_name, self._func_wrapped)

    def sfd_ebv_func(self, ra, dec):
        lb = radec_to_lb(ra, dec, degree=True)
        l, b = lb[:, 0], lb[:, 1]
        return self.sfd(l, b, np.ones_like(l))



import subprocess
from . import (
    downloader,
    _GAIA_DR3_GAIASOURCE_PARENT,
    _GAIA_DR3_ASTROPHYS_PARENT,
    _GAIA_DR3_ALLWISE_NEIGHBOUR_PARENT,
    _GAIA_DR3_2MASS_NEIGHBOUR_PARENT,
    _GAIA_DR3_XP_CONTINUOUS_PARENT,
    _GAIA_DR3_XP_SAMPLED_PARENT,
    _GAIA_DR3_RVS_PARENT,
    _ALLWISE_PARENT,
    _2MASS_PARENT,
)


def download_gaia_source(test=False):
    _GAIA_DR3_GAIASOURCE_PARENT.mkdir(parents=True, exist_ok=True)
    _url = "http://cdn.gea.esac.esa.int/Gaia/gdr3/gaia_source/"
    if test:
        downloader(f"{_url}GaiaSource_000000-003111.csv.gz", _GAIA_DR3_GAIASOURCE_PARENT.joinpath("GaiaSource_000000-003111.csv.gz"), "test", test=test)
        downloader(f"{_url}GaiaSource_003112-005263.csv.gz", _GAIA_DR3_GAIASOURCE_PARENT.joinpath("GaiaSource_003112-005263.csv.gz"), "test", test=test)
        # cmd_str = f"cd {_GAIA_DR3_GAIASOURCE_PARENT.as_posix()} && curl --no-clobber -s -O {_url}GaiaSource_000000-003111.csv.gz"
        # subprocess.run(cmd_str, shell=True)
        # cmd_str = f"cd {_GAIA_DR3_GAIASOURCE_PARENT.as_posix()} && curl --no-clobber -s -O {_url}GaiaSource_003112-005263.csv.gz"
        # subprocess.run(cmd_str, shell=True)
    else:
        cmd_str = f"wget -P {_GAIA_DR3_GAIASOURCE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent --recursive --level=1 --no-directories {_url}"
        subprocess.run(cmd_str, shell=True)


def download_gaia_astrophysical_parameters(test=False):
    _GAIA_DR3_ASTROPHYS_PARENT.mkdir(parents=True, exist_ok=True)
    _url = "http://cdn.gea.esac.esa.int/Gaia/gdr3/Astrophysical_parameters/astrophysical_parameters/"
    if test:
        downloader(f"{_url}AstrophysicalParameters_000000-003111.csv.gz", _GAIA_DR3_ASTROPHYS_PARENT.joinpath("AstrophysicalParameters_000000-003111.csv.gz"), "test", test=test)
        downloader(f"{_url}AstrophysicalParameters_003112-005263.csv.gz", _GAIA_DR3_ASTROPHYS_PARENT.joinpath("AstrophysicalParameters_003112-005263.csv.gz"), "test", test=test)
        # cmd_str = f"cd {_GAIA_DR3_ASTROPHYS_PARENT.as_posix()} && curl --no-clobber -s -O {_url}AstrophysicalParameters_000000-003111.csv.gz"
        # subprocess.run(cmd_str, shell=True)
        # cmd_str = f"cd {_GAIA_DR3_ASTROPHYS_PARENT.as_posix()} && curl --no-clobber -s -O {_url}AstrophysicalParameters_003112-005263.csv.gz"
        # subprocess.run(cmd_str, shell=True)
    else:
        cmd_str = f"wget -P {_GAIA_DR3_ASTROPHYS_PARENT.as_posix()} --no-clobber --no-verbose --no-parent --recursive --level=1 --no-directories {_url}"
        subprocess.run(cmd_str, shell=True)


def download_allwise_best_neightbour(test=False):
    _GAIA_DR3_ALLWISE_NEIGHBOUR_PARENT.mkdir(parents=True, exist_ok=True)
    _url = "http://cdn.gea.esac.esa.int/Gaia/gedr3/cross_match/allwise_best_neighbour/"
    if test:
        downloader(f"{_url}allwiseBestNeighbour0001.csv.gz", _GAIA_DR3_ALLWISE_NEIGHBOUR_PARENT.joinpath("allwiseBestNeighbour0001.csv.gz"), "test", test=test)
        downloader(f"{_url}allwiseBestNeighbour0002.csv.gz", _GAIA_DR3_ALLWISE_NEIGHBOUR_PARENT.joinpath("allwiseBestNeighbour0002.csv.gz"), "test", test=test)
        # cmd_str = f"cd {_GAIA_DR3_ALLWISE_NEIGHBOUR_PARENT.as_posix()} && curl --no-clobber -s -O {_url}allwiseBestNeighbour0001.csv.gz"
        # subprocess.run(cmd_str, shell=True)
        # cmd_str = f"cd {_GAIA_DR3_ALLWISE_NEIGHBOUR_PARENT.as_posix()} && curl --no-clobber -s -O {_url}allwiseBestNeighbour0002.csv.gz"
        # subprocess.run(cmd_str, shell=True)
    else:
        cmd_str = f"wget -P {_GAIA_DR3_ALLWISE_NEIGHBOUR_PARENT.as_posix()} --no-clobber --no-verbose --no-parent --recursive --level=1 --no-directories {_url}"
        subprocess.run(cmd_str, shell=True)


def download_2mass_best_neightbour(test=False):
    _GAIA_DR3_2MASS_NEIGHBOUR_PARENT.mkdir(parents=True, exist_ok=True)
    _url = (
        "http://cdn.gea.esac.esa.int/Gaia/gedr3/cross_match/tmasspscxsc_best_neighbour/"
    )
    if test:
        downloader(f"{_url}tmasspscxscBestNeighbour0001.csv.gz", _GAIA_DR3_2MASS_NEIGHBOUR_PARENT.joinpath("tmasspscxscBestNeighbour0001.csv.gz"), "test", test=test)
        downloader(f"{_url}tmasspscxscBestNeighbour0002.csv.gz", _GAIA_DR3_2MASS_NEIGHBOUR_PARENT.joinpath("tmasspscxscBestNeighbour0002.csv.gz"), "test", test=test)
        # cmd_str = f"cd {_GAIA_DR3_2MASS_NEIGHBOUR_PARENT.as_posix()} && curl --no-clobber -s -O {_url}tmasspscxscBestNeighbour0001.csv.gz"
        # subprocess.run(cmd_str, shell=True)
        # cmd_str = f"cd {_GAIA_DR3_2MASS_NEIGHBOUR_PARENT.as_posix()} && curl --no-clobber -s -O {_url}tmasspscxscBestNeighbour0002.csv.gz"
        # subprocess.run(cmd_str, shell=True)
    else:
        cmd_str = f"wget -P {_GAIA_DR3_2MASS_NEIGHBOUR_PARENT.as_posix()} --no-clobber --no-verbose --no-parent --recursive --level=1 --no-directories {_url}"
        subprocess.run(cmd_str, shell=True)

def download_2mass(test=False):
    _2MASS_PARENT.mkdir(parents=True, exist_ok=True)
    _url = (
        "https://irsa.ipac.caltech.edu/2MASS/download/allsky/"
    )
    if test:
        downloader(f"{_url}psc_aaa.gz", _2MASS_PARENT.joinpath("psc_aaa.gz"), "test", test=test)
        downloader(f"{_url}psc_aab.gz", _2MASS_PARENT.joinpath("psc_aab.gz"), "test", test=test)
        # cmd_str = f"cd {_2MASS_PARENT.as_posix()} && curl --no-clobber -s -O {_url}psc_aaa.gz"
        # subprocess.run(cmd_str, shell=True)
        # cmd_str = f"cd {_2MASS_PARENT.as_posix()} && curl --no-clobber -s -O {_url}psc_aab.gz"
        # subprocess.run(cmd_str, shell=True)
    else:
        cmd_str = f"wget -P {_2MASS_PARENT.as_posix()} --no-clobber --no-verbose --no-parent --recursive --level=1 --no-directories {_url}"
        subprocess.run(cmd_str, shell=True)


def download_allwise(test=False):
    _ALLWISE_PARENT.mkdir(parents=True, exist_ok=True)
    _url = "http://irsa.ipac.caltech.edu/data/download/wise-allwise/"
    # https://irsa.ipac.caltech.edu/data/download/wise-allwise/wget_bz2.script
    cmd_list = [
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/README.txt",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat.bz2.sizes.txt",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat.bz2.md5.txt",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part01.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part02.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part03.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part04.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part05.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part06.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part07.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part08.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part09.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part10.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part11.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part12.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part13.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part14.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part15.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part16.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part17.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part18.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part19.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part20.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part21.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part22.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part23.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part24.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part25.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part26.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part27.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part28.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part29.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part30.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part31.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part32.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part33.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part34.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part35.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part36.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part37.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part38.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part39.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part40.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part41.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part42.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part43.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part44.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part45.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part46.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part47.bz2",
        f"wget -P {_ALLWISE_PARENT.as_posix()} --no-clobber --no-verbose --no-parent http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part48.bz2",
    ]

    if test:
        downloader(f"http://irsa.ipac.caltech.edu/data/download/wise-allwise/wise-allwise-cat-part01.bz2", _ALLWISE_PARENT.joinpath("wise-allwise-cat-part01.bz2"), "test", test=test)
        # subprocess.run(cmd_list[3], shell=True)
    else:
        for cmd_str in cmd_list:
            subprocess.run(cmd_str, shell=True)


def download_gaia_xp_continuous(test=False):
    _GAIA_DR3_XP_CONTINUOUS_PARENT.mkdir(parents=True, exist_ok=True)
    _url = "http://cdn.gea.esac.esa.int/Gaia/gdr3/Spectroscopy/xp_continuous_mean_spectrum/"
    if test:
        downloader(f"{_url}XpContinuousMeanSpectrum_000000-003111.csv.gz", _GAIA_DR3_XP_CONTINUOUS_PARENT.joinpath("XpContinuousMeanSpectrum_000000-003111.csv.gz"), "test", test=test)
        # cmd_str = f"cd {_GAIA_DR3_XP_CONTINUOUS_PARENT.as_posix()} && curl --no-clobber -s -O {_url}XpContinuousMeanSpectrum_000000-003111.csv.gz"
        # subprocess.run(cmd_str, shell=True)
    else:
        cmd_str = f"wget -P {_GAIA_DR3_XP_CONTINUOUS_PARENT.as_posix()} --no-clobber --no-verbose --no-parent --recursive --level=1 --no-directories {_url}"
        subprocess.run(cmd_str, shell=True)


def download_gaia_xp_sampled(test=False):
    _GAIA_DR3_XP_SAMPLED_PARENT.mkdir(parents=True, exist_ok=True)
    _url = "http://cdn.gea.esac.esa.int/Gaia/gdr3/Spectroscopy/xp_sampled_mean_spectrum/"
    if test:
        downloader(f"{_url}XpSampledMeanSpectrum_000000-003111.csv.gz", _GAIA_DR3_XP_SAMPLED_PARENT.joinpath("XpSampledMeanSpectrum_000000-003111.csv.gz"), "test", test=test)
        # cmd_str = f"cd {_GAIA_DR3_XP_SAMPLED_PARENT.as_posix()} && curl --no-clobber -s -O {_url}XpSampledMeanSpectrum_000000-003111.csv.gz"
        # subprocess.run(cmd_str, shell=True)
    else:
        cmd_str = f"wget -P {_GAIA_DR3_XP_SAMPLED_PARENT.as_posix()} --no-clobber --no-verbose --no-parent --recursive --level=1 --no-directories {_url}"
        subprocess.run(cmd_str, shell=True)


def download_gaia_rvs(test=False):
    _GAIA_DR3_RVS_PARENT.mkdir(parents=True, exist_ok=True)
    _url = "http://cdn.gea.esac.esa.int/Gaia/gdr3/Spectroscopy/rvs_mean_spectrum/"
    if test:
        downloader(f"{_url}RvsMeanSpectrum_000000-003111.csv.gz", _GAIA_DR3_RVS_PARENT.joinpath("RvsMeanSpectrum_000000-003111.csv.gz"), "test", test=test)
        # cmd_str = f"cd {_GAIA_DR3_RVS_PARENT.as_posix()} && curl --no-clobber -s -O {_url}RvsMeanSpectrum_000000-003111.csv.gz"
        # subprocess.run(cmd_str, shell=True)
    else:
        cmd_str = f"wget -P {_GAIA_DR3_RVS_PARENT.as_posix()} --no-clobber --no-verbose --no-parent --recursive --level=1 --no-directories {_url}"
        subprocess.run(cmd_str, shell=True)



import os
import pathlib
from pkg_resources import get_distribution

version = __version__ = get_distribution("MyGaiaDB").version

# make sure (shared) database folder exists
astro_data_path = os.getenv("MY_ASTRO_DATA")
if astro_data_path is None:
    raise EnvironmentError("Please specify an environment variable - MY_ASTRO_DATA")
else:
    astro_data_path = pathlib.Path(astro_data_path).expanduser()
    if not astro_data_path.exists():
        astro_data_path.mkdir()

# make sure user-specific database folder exists
mygaiadb_folder = pathlib.Path.home().joinpath(".mygaiadb")
mygaiadb_default_db = mygaiadb_folder.joinpath("mygaiadb.db")
mygaiadb_usertable_db = mygaiadb_folder.joinpath("user_table.db")
if not mygaiadb_folder.exists():
    mygaiadb_folder.mkdir()
mygaiadb_default_db.touch()
mygaiadb_usertable_db.touch()

gaia_sql_db_path = astro_data_path.joinpath("gaia_mirror", "gaiadr3.db")
gaia_astro_param_sql_db_path = astro_data_path.joinpath("gaia_mirror", "gaiadr3_astrophysical_params.db")
gaia_xp_coeff_h5_path = astro_data_path.joinpath("gaia_mirror" ,"xp_continuous_mean_spectrum_allinone.h5")
tmass_sql_db_path = astro_data_path.joinpath("2mass_mirror", "tmass.db")
allwise_sql_db_path = astro_data_path.joinpath("allwise_mirror", "allwise.db")



import numpy as np
import astropy.units as u


def SameSide(A, R):
    """Place the angle into same quadrant as referant"""
    A_rad = A / 180 * np.pi
    R_rad = R / 180 * np.pi
    return np.where(
        (
            ((np.sign(np.cos(A_rad))) == (np.sign(np.cos(R_rad)))) == ((np.sign(np.sin(A_rad))) == (np.sign(np.sin(R_rad))))
        ),
        (A + 36000) % 360,
        (A + 36180) % 360,
    )


def radec_to_ecl(ra, dec):
    """
    refers to section 1.5.3 in https://www.cosmos.esa.int/documents/532822/552851/vol1_all.pdf

    this relation is good for Gaia with epoch 2016, accurate to almost machine precision
    """
    e = 23.43928083333333 / 180 * np.pi
    ra_rad = np.asarray(ra) / 180. * np.pi + (0.05542 * u.arcsec).to(u.rad).value
    dec_rad = np.asarray(dec) / 180. * np.pi
    l = np.arctan(
        np.tan(ra_rad) * np.cos(e) + np.tan(dec_rad) * np.sin(e) / np.cos(ra_rad)
    ) / (np.pi / 180.)
    b = np.arcsin(
        np.sin(dec_rad) * np.cos(e) - np.cos(dec_rad) * np.sin(e) * np.sin(ra_rad)
    ) / (np.pi / 180.)
    l = SameSide(l, ra_rad)
    return l, b



import os
import re
import sys
import stat
import ctypes
import pathlib
import inspect
import sqlite3
import sysconfig
import contextlib
import numpy as np
import pandas as pd
from tqdm import tqdm
from mygaiadb import (
    mygaiadb_default_db,
    mygaiadb_usertable_db,
    gaia_sql_db_path,
    tmass_sql_db_path,
    allwise_sql_db_path,
    __version__
)


class QueryCallback:
    """
    Callback to add new column to SQL query on the fly
    """

    def __init__(self, new_col_name, func):
        """
        INPUT:
            new_col_name (string): Name of the new column you wan to add
            func (function): function that maps query columns to new columns, arguements of this function need to have \
                the same names as columns in query
        """
        self.new_col_name = new_col_name
        self.func = func
        self.required_col = list(inspect.getfullargspec(self.func))[0]


class LocalGaiaSQL:
    """
    Class for local Gaia SQL database
    """

    def __init__(
        self,
        load_tmass=True,
        load_allwise=True,
        load_ext=True,
        readonly_guard=True,
    ):
        """
        Parameters
        ----------
        load_tmass : bool
            whether to load 2mass table
        load_allwise : bool
            whether to load allwise table
        load_ext : bool
            whether to load sqlite extension
        readonly_guard : bool
            whether to ensure the databases are read-only
        """
        self.load_tmass = load_tmass
        self.load_allwise = load_allwise
        self.load_ext = load_ext
        self.readonly_guard = readonly_guard
        self.attached_db_name = []

        # flag for windows or not
        self.win32 = sys.platform.startswith("win32")

        self.conn, self.cursor = self._load_db()

        # ipython Auto-completion
        try:
            from IPython import get_ipython
            def name_completer(ipython, event):
                out = self.list_all_tables()
                out.extend(list(f"user_table.{i}" for i in self.list_user_tables().keys()))
                return out
            get_ipython().set_hook("complete_command", name_completer,
                                   re_key=".*get_table_column")
        except:
            pass

    def _check_callbacks_header(self, headers, callbacks):
        """
        Helper function to check if all columns required by all callbacks are presented in query

        Parameters
        ----------
        headers : list of string
            SQL query result header
        callbacks : QueryCallback object
            callbacks used in a query
        """
        for i in callbacks:
            for j in i.required_col:
                if not j in headers:
                    raise NameError(
                        f"Callback for new column {i.new_col_name} requires column {j} but not presented in your query"
                    )

    def _result_after_callbacks(self, df, callbacks):
        for i in callbacks:
            func_dist = {}
            for j in i.required_col:
                func_dist[j] = df[j]
            df[i.new_col_name] = i.func(**func_dist)
        return df

    def _read_only(self, file_path):
        # set read only premission for all loaded dataset to prevent accidental change
        if self.win32:
            os.chmod(file_path, stat.S_IREAD)
        else:
            os.chmod(file_path, 0o444)

    def _file_exist(self, path):
        if not os.path.exists(path):
            raise FileExistsError(f"Database at {path} does not exist")

    def preprocess_query(func):
        """
        Decoractor to preprocess query. Mostly to increase compatability with Gaia ADQL
        """

        def preprocess_logic(self, *args, **kwargs):
            # either it is keyworded or the first arguement
            if "query" in kwargs:
                query = kwargs["query"]
            else:
                query = args[0]

            # turn TOP * to LIMIT *
            m = re.search(r"TOP (\d+)", query, re.IGNORECASE)
            if m is not None:
                limit_num = m.group(1)
                query = re.sub(r"\s+[TOP]+\s+[0-9]+", "", query, flags=re.IGNORECASE)
                query += f""" LIMIT {limit_num}"""

            # turn true to 1
            query = re.sub(r"[']\s?\b(t)\b\s?[']", "1 ", query, flags=re.IGNORECASE)
            query = re.sub(r"[']\s?\b(f)\b\s?[']", "0 ", query, flags=re.IGNORECASE)
            query = re.sub(r"[']\s?\b(true)\b\s?[']", "1 ", query, flags=re.IGNORECASE)
            query = re.sub(r"[']\s?\b(false)\b\s?[']", "0 ", query, flags=re.IGNORECASE)
            if "query" in kwargs:
                kwargs["query"] = query
            else:
                args = list(args)
                args[0] = query
                args = tuple(args)
            return func(self, *args, **kwargs)

        return preprocess_logic

    def _load_db(self):
        """
        Get SQL database connection and cursor used in this work for Gaia DR3 as the main table; 2MASS, ALLWISE and gaia astrophysical parameters as virtual tables
        """
        self._file_exist(mygaiadb_default_db)
        conn = sqlite3.connect(gaia_sql_db_path)
        conn.create_function("mygaiadb_version", 0, lambda: __version__, deterministic=True)
        c = conn.cursor()
        if self.load_ext:
            self._load_sqlite3_ext(conn)

        # ======================= must load table =======================
        self._file_exist(gaia_sql_db_path)
        if self.readonly_guard:
            self._read_only(gaia_sql_db_path)  # set read-only before loading it
        c.execute(f"""ATTACH DATABASE '{gaia_sql_db_path}' AS gaiadr3""")
        self.attached_db_name.append("gaiadr3")
        self._file_exist(mygaiadb_usertable_db)
        c.execute(
            f"""ATTACH DATABASE '{mygaiadb_usertable_db}' AS user_table"""
        )  # don't read-only
        # ======================= must load table =======================

        # ======================= optional table =======================
        if self.load_tmass:
            self._file_exist(tmass_sql_db_path)
            if self.readonly_guard:
                self._read_only(tmass_sql_db_path)  # set read-only before loading it
            c.execute(f"""ATTACH DATABASE '{tmass_sql_db_path}' AS tmass""")
            self.attached_db_name.append("tmass")
        if self.load_allwise:
            self._file_exist(allwise_sql_db_path)
            if self.readonly_guard:
                self._read_only(allwise_sql_db_path)  # set read-only before loading it
            c.execute(f"""ATTACH DATABASE '{allwise_sql_db_path}' AS allwise""")
            self.attached_db_name.append("allwise")
        # ======================= optional table =======================
        return conn, c
    
    @staticmethod
    def _load_sqlite3_ext(c):
        c.enable_load_extension(True)
        # Find and load the library
        _lib = None
        _libname = ctypes.util.find_library("astroqlite_c")
        _ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")
        if _libname:
            _lib = _libname
        if _lib is None:
            for path in sys.path:
                _temp_lib = pathlib.Path(path).joinpath(f"astroqlite_c{_ext_suffix}")
                if _temp_lib.exists():
                    _lib = _temp_lib.as_posix()
                    c.load_extension(_lib)
                    break
        if _lib is None:
            raise IOError("MyGaiaDB SQL C extension not found")

    @preprocess_query
    def save_csv(
        self, query, filename, chunchsize=50000, overwrite=True, callbacks=None
    ):
        """
        Given query, save the fetchall() result to csv, "chunchsize" number of rows at each time until finished

        Parameters
        ----------
        query : string
            Query string
        filename : string
            filename (*.csv) to be saved
        Returns
        -------
        None
        """
        if callbacks is not None and not isinstance(callbacks, list):
            raise TypeError("callbacks must be a list")

        self.cursor.execute(query)
        if os.path.exists(filename) and not overwrite:
            raise SystemError(f"{os.path.abspath(filename)} already existed!")
        # write header rows
        header_og = [d[0] for d in self.cursor.description]
        if callbacks is not None:
            header_big = [d[0] for d in self.cursor.description]
            header_big.extend([i.new_col_name for i in callbacks])
            self._check_callbacks_header(header_og, callbacks)
        else:
            header_big = header_og
        pd.DataFrame(columns=header_big).to_csv(filename, index=False)
        # csv_out.writerow(header)
        first_flag = True
        with tqdm(unit=" rows") as pbar:
            pbar.set_description_str("Rows written: ")
            while True:  # looping until the end
                results = self.cursor.fetchmany(chunchsize)
                if results == []:
                    break
                _df = pd.DataFrame(results, columns=header_og)
                if callbacks is not None:
                    _df = self._result_after_callbacks(_df, callbacks)
                _df.to_csv(
                    filename,
                    mode="a" if not first_flag else "w",
                    index=False,
                    header=False if not first_flag else True,
                )
                first_flag = False
                pbar.update(len(_df))
        return None

    @preprocess_query
    def query(self, query, callbacks=None):
        """
        Get result from query to pandas dataframe, ONLY USE THIS FOR SMALL QUERY

        Parameters
        ----------
        query : string
            Query string

        Returns
        -------
        df: pandas.Dataframe
        """
        _df = pd.read_sql_query(query, self.conn)
        if callbacks is not None:
            self._check_callbacks_header(_df.columns, callbacks)
            _df = self._result_after_callbacks(_df, callbacks)
        return _df

    @preprocess_query
    def execution_plan(self, query):
        """
        Get execution plan for a query, to debug to improve indexing
        """
        query = (
            """
        EXPLAIN QUERY PLAN
        """
            + query
        )
        self.cursor.execute(query)
        return self.cursor.fetchall()

    def xmatch(self, user_id, db_id, query):
        """
        cross-matching given a list of source_id
        """
        raise NotImplementedError()

    def upload_user_table(self, df, tablename=None):
        """
        Add a custom user table

        Parameters
        ----------
        df : pandas.Dataframe
            pandas dataframe to be added
        tablename: str
            Table name
        """
        if tablename is None:
            raise NameError("You need to specify a name for the new table")
        with contextlib.closing(sqlite3.connect(mygaiadb_usertable_db)) as conn:
            df.to_sql(f"{tablename}", conn, if_exists="fail", index=False)

    def remove_user_table(self, tablename=None, reclaim=False):
        """
        Remove a custom user table

        Parameters
        ----------
        tablename: str
            Table name
        reclaim: bool
            Whether to reclaim disk space after removing a table
        """
        if tablename is None:
            raise NameError("You need to specify a name for the table to be removed")
        with contextlib.closing(sqlite3.connect(mygaiadb_usertable_db)) as conn:
            conn.execute(f"""DROP TABLE {tablename}""")
            if reclaim:
                conn.execute("""VACUUM""")

    def list_user_tables(self):
        """
        Get the list of user table name

        Returns
        -------
        result: dict
            dictionary of all user table name and columns
        """
        self.cursor.execute(
            """
        SELECT name
        FROM user_table.sqlite_schema
        WHERE type ='table'
        """
        )
        a = self.cursor.fetchall()
        result = {}

        # TODO: add table length??
        for i in a:
            name = i[0]
            cols = self.get_table_column(name=f"user_table.{name}")
            result[name] = cols

        return result

    def list_all_tables(self):
        """
        Get the list of all table name

        Returns
        -------
        result: list
            list of tables with the format of DATABASE_NAME.TABLE_NAME
        """
        result = []
        for i in self.attached_db_name:
            self.cursor.execute(
                f"""
            SELECT name FROM {i}.sqlite_schema
            WHERE type ='table'
            """
            )
            a = self.cursor.fetchall()
            result.extend(list(f"{i}.{_a[0]}" for _a in a))
        return result

    def get_table_column(self, name=None):
        """
        Get the list of column from a table

        Parameters
        ----------
        name: str
            Table name with the format of DATABASE_NAME.TABLE_NAME

        Returns
        -------
        result: list
            list of columns of DATABASE_NAME.TABLE_NAME
        """
        if not "." in name:
            raise NameError(
                "Table name need to be with the format of DATABASE_NAME.TABLE_NAME"
            )
        result = pd.read_sql_query(f"""SELECT * FROM {name} LIMIT 1""", self.conn)
        return [i for i in result.columns]



import tqdm
import shutil
import pathlib
import requests
from .. import astro_data_path

_GAIA_PARENT = astro_data_path.joinpath("gaia_mirror")
_GAIA_DR3_PARENT = _GAIA_PARENT.joinpath("Gaia", "gdr3")
_GAIA_DR3_GAIASOURCE_PARENT = _GAIA_DR3_PARENT.joinpath("gaia_source")
_GAIA_DR3_ASTROPHYS_PARENT = _GAIA_DR3_PARENT.joinpath("Astrophysical_parameters", "astrophysical_parameters")
_GAIA_DR3_ALLWISE_NEIGHBOUR_PARENT = _GAIA_DR3_PARENT.joinpath("cross_match", "allwise_best_neighbour")
_GAIA_DR3_2MASS_NEIGHBOUR_PARENT = _GAIA_DR3_PARENT.joinpath("cross_match", "tmasspscxsc_best_neighbour")
_GAIA_DR3_XP_CONTINUOUS_PARENT = _GAIA_DR3_PARENT.joinpath("Spectroscopy", "xp_continuous_mean_spectrum")
_GAIA_DR3_XP_SAMPLED_PARENT = _GAIA_DR3_PARENT.joinpath("Spectroscopy", "xp_sampled_mean_spectrum")
_GAIA_DR3_RVS_PARENT = _GAIA_DR3_PARENT.joinpath("Spectroscopy", "rvs_mean_spectrum")
_2MASS_PARENT = astro_data_path.joinpath("2mass_mirror")
_ALLWISE_PARENT = astro_data_path.joinpath("allwise_mirror")


def downloader(url, fullfilename, name, test=False, session=None):
    """
    url: URL of data file
    fullfilename: full local path
    name: name of the task
    session: Requests session
    """
    if session is None:
        s = requests.Session()
    user_agent = "Mozilla/5.0"
    r = requests.get(url, stream=True, allow_redirects=True, verify=True, headers={"User-Agent": user_agent})
    if r.status_code == 404:
        raise ConnectionError(f"Cannot find {name} data file at {url}")
    r.raise_for_status()  # Will only raise for 4xx codes\
    path = pathlib.Path(fullfilename).expanduser().resolve()
    path.parent.mkdir(parents=True, exist_ok=True)
    if not test or not path.exists():
        file_size = int(r.headers.get('Content-Length', 0))
        # r.raw.read
        with tqdm.tqdm.wrapattr(r.raw, "read", total=file_size, desc=f"Download {name}") as r_raw:
            with path.open("wb") as f:
                shutil.copyfileobj(r_raw, f)



import h5py
import tqdm
import numpy as np
from .. import astro_data_path, gaia_xp_coeff_h5_path


def yield_xp_coeffs(source_ids, assume_unique=True, rdcc_nbytes=16*1024**3, rdcc_nslots=10e7):
    """
    Function to yield XP coeffs according to their healpixs from source_id

    Parameters
    ----------
    source_ids: (int, list, ndarray)
        source id
    assume_unique: bool
        whether to assume the list of source id is unique
    rdcc_nbytes: int
        h5py cache in bytes
    rdcc_nslots: int
        h5py cache number of slots
    """
    reduced_source_ids = source_ids // 8796093022208
    total_num = len(source_ids)

    h5f = h5py.File(gaia_xp_coeff_h5_path, "r", rdcc_nbytes=rdcc_nbytes, rdcc_nslots=rdcc_nslots)
    file_names = list(h5f.keys())

    # Extract HEALPix level-8 from file name
    healpix_8_min = [
        int(file[0:file.rfind("-")]) for file in file_names
    ]
    healpix_8_max = [
        int(file[file.rfind("-") + 1:]) for file in file_names
    ]
    reference_file = {
        "file": file_names,
        "healpix8_min": np.asarray(healpix_8_min),
        "healpix8_max": np.asarray(healpix_8_max),
    }
    for i in tqdm.tqdm(range(len(reference_file["healpix8_min"]))):
        good_idx = (reference_file["healpix8_min"][i] <= reduced_source_ids) & (reduced_source_ids <= reference_file["healpix8_max"][i])
        current_source_ids = source_ids[good_idx]
        spec_f = h5f[f"{reference_file['file'][i]}"]
        matches, idx1, idx2 = np.intersect1d(
            current_source_ids,
            spec_f["source_id"][()],
            assume_unique=assume_unique,
            return_indices=True,
        )
        bp_coeffs = spec_f["bp_coefficients"][()]
        rp_coeffs = spec_f["rp_coefficients"][()]

        if len(matches) > 0:
            coeffs = np.hstack([bp_coeffs[idx2], rp_coeffs[idx2]])
            yield coeffs, np.arange(total_num)[good_idx][idx1]
        else:
            # TODO: I think maybe should not pass, need more test
            pass



import sqlite3
from pathlib import Path
import os
import glob
import tqdm
import h5py
import numpy as np
import pandas as pd
from . import (
    _GAIA_DR3_GAIASOURCE_PARENT,
    _GAIA_DR3_ASTROPHYS_PARENT,
    _GAIA_DR3_ALLWISE_NEIGHBOUR_PARENT,
    _GAIA_DR3_2MASS_NEIGHBOUR_PARENT,
    _2MASS_PARENT,
    _ALLWISE_PARENT
)
from .. import astro_data_path, gaia_sql_db_path, tmass_sql_db_path, allwise_sql_db_path, gaia_xp_coeff_h5_path


def compile_xp_continuous_allinone_h5():
    base_path = astro_data_path.joinpath(
        "gaia_mirror",
        "Gaia",
        "gdr3",
        "Spectroscopy",
        "xp_continuous_mean_spectrum",
    )
    file_paths = list(base_path.glob("*.h5"))

    if len(file_paths) == 0:
        raise FileNotFoundError(f"Gaia data does not exist at {base_path}")

    file_names = [x.name for x in file_paths]
    healpix_8_min = [
        int(file[file.find("_") + 1 : file.rfind("-")]) for file in file_names
    ]
    healpix_8_max = [
        int(file[file.rfind("-") + 1 : file.rfind(".h5")]) for file in file_names
    ]

    if gaia_xp_coeff_h5_path.exists():
        raise FileExistsError(f"File already existed in {gaia_xp_coeff_h5_path}")

    h5f = h5py.File(gaia_xp_coeff_h5_path, "w")
    for i in tqdm.tqdm(range(len(file_paths))):
        temp_h5_data = h5py.File(file_paths[i].as_posix(), "r")
        gp = h5f.create_group(f"{healpix_8_min[i]}-{healpix_8_max[i]}")
        gp.create_dataset("source_id", data=temp_h5_data["source_id"][()])
        gp.create_dataset("solution_id", data=temp_h5_data["solution_id"][()])
        gp.create_dataset(
            "bp_basis_function_id", data=temp_h5_data["bp_basis_function_id"][()]
        )
        gp.create_dataset(
            "bp_degrees_of_freedom", data=temp_h5_data["bp_degrees_of_freedom"][()]
        )
        gp.create_dataset("bp_n_parameters", data=temp_h5_data["bp_n_parameters"][()])
        gp.create_dataset(
            "bp_n_measurements", data=temp_h5_data["bp_n_measurements"][()]
        )
        gp.create_dataset(
            "bp_n_rejected_measurements",
            data=temp_h5_data["bp_n_rejected_measurements"][()],
        )
        gp.create_dataset(
            "bp_standard_deviation", data=temp_h5_data["bp_standard_deviation"][()]
        )
        gp.create_dataset("bp_chi_squared", data=temp_h5_data["bp_chi_squared"][()])
        gp.create_dataset(
            "bp_coefficient_errors", data=temp_h5_data["bp_coefficient_errors"][()]
        )
        gp.create_dataset("bp_coefficients", data=temp_h5_data["bp_coefficients"][()])
        gp.create_dataset(
            "bp_n_relevant_bases", data=temp_h5_data["bp_n_relevant_bases"][()]
        )
        gp.create_dataset(
            "bp_relative_shrinking", data=temp_h5_data["bp_relative_shrinking"][()]
        )

        gp.create_dataset(
            "rp_basis_function_id", data=temp_h5_data["rp_basis_function_id"][()]
        )
        gp.create_dataset(
            "rp_degrees_of_freedom", data=temp_h5_data["rp_degrees_of_freedom"][()]
        )
        gp.create_dataset("rp_n_parameters", data=temp_h5_data["rp_n_parameters"][()])
        gp.create_dataset(
            "rp_n_measurements", data=temp_h5_data["rp_n_measurements"][()]
        )
        gp.create_dataset(
            "rp_n_rejected_measurements",
            data=temp_h5_data["rp_n_rejected_measurements"][()],
        )
        gp.create_dataset(
            "rp_standard_deviation", data=temp_h5_data["rp_standard_deviation"][()]
        )
        gp.create_dataset("rp_chi_squared", data=temp_h5_data["rp_chi_squared"][()])
        gp.create_dataset(
            "rp_coefficient_errors", data=temp_h5_data["rp_coefficient_errors"][()]
        )
        gp.create_dataset("rp_coefficients", data=temp_h5_data["rp_coefficients"][()])
        gp.create_dataset(
            "rp_n_relevant_bases", data=temp_h5_data["rp_n_relevant_bases"][()]
        )
        gp.create_dataset(
            "rp_relative_shrinking", data=temp_h5_data["rp_relative_shrinking"][()]
        )
        temp_h5_data.close()
    h5f.close()


def compile_xp_continuous_h5(save_correlation_matrix=False):
    for i_path in tqdm.tqdm(
        list(astro_data_path.joinpath(
            "Gaia", "gdr3", "Spectroscopy", "xp_continuous_mean_spectrum"
        ).glob("*.csv.gz")),
        desc="XP coeffs",
    ):
        file_path_f = ascii.read(i_path)

        bp_basis_function_id = np.asarray(
            file_path_f["bp_basis_function_id"], dtype=np.int16
        )
        bp_degrees_of_freedom = np.asarray(
            file_path_f["bp_degrees_of_freedom"], dtype=np.int16
        )
        bp_n_parameters = np.asarray(file_path_f["bp_n_parameters"], dtype=np.int8)
        bp_n_measurements = np.asarray(file_path_f["bp_n_measurements"], dtype=np.int16)
        bp_n_rejected_measurements = np.asarray(
            file_path_f["bp_n_rejected_measurements"], dtype=np.int16
        )
        bp_standard_deviation = np.asarray(
            file_path_f["bp_standard_deviation"], dtype=np.float32
        )
        bp_chi_squared = np.asarray(file_path_f["bp_chi_squared"], dtype=np.float32)
        if save_correlation_matrix:
            bp_coefficient_correlations = np.stack(
                file_path_f["bp_coefficient_correlations"]
            )
        bp_coefficient_errors = np.stack(file_path_f["bp_coefficient_errors"])
        bp_coefficients = np.stack(file_path_f["bp_coefficients"])
        bp_n_relevant_bases = np.asarray(
            file_path_f["bp_n_relevant_bases"], dtype=np.int16
        )
        bp_relative_shrinking = np.asarray(
            file_path_f["bp_relative_shrinking"], dtype=np.float32
        )

        rp_basis_function_id = np.asarray(
            file_path_f["rp_basis_function_id"], dtype=np.int16
        )
        rp_degrees_of_freedom = np.asarray(
            file_path_f["rp_degrees_of_freedom"], dtype=np.int16
        )
        rp_n_parameters = np.asarray(file_path_f["rp_n_parameters"], dtype=np.int8)
        rp_n_measurements = np.asarray(file_path_f["rp_n_measurements"], dtype=np.int16)
        rp_n_rejected_measurements = np.asarray(
            file_path_f["rp_n_rejected_measurements"], dtype=np.int16
        )
        rp_standard_deviation = np.asarray(
            file_path_f["rp_standard_deviation"], dtype=np.float32
        )
        rp_chi_squared = np.asarray(file_path_f["rp_chi_squared"], dtype=np.float32)
        if save_correlation_matrix:
            rp_coefficient_correlations = np.stack(
                file_path_f["rp_coefficient_correlations"]
            )
        rp_coefficient_errors = np.stack(file_path_f["rp_coefficient_errors"])
        rp_coefficients = np.stack(file_path_f["rp_coefficients"])
        rp_n_relevant_bases = np.asarray(
            file_path_f["rp_n_relevant_bases"], dtype=np.int16
        )
        rp_relative_shrinking = np.asarray(
            file_path_f["rp_relative_shrinking"], dtype=np.float32
        )

        file_names_wo_ext = i_path.name[:-7]
        with h5py.File(
            f"./xp_continuous_mean_spectrum/{file_names_wo_ext}.h5", "w"
        ) as h5f:
            h5f.create_dataset("source_id", data=file_path_f["source_id"].data)
            h5f.create_dataset("solution_id", data=file_path_f["solution_id"].data)
            h5f.create_dataset("bp_basis_function_id ", data=bp_basis_function_id)
            h5f.create_dataset("bp_degrees_of_freedom ", data=bp_degrees_of_freedom)
            h5f.create_dataset("bp_n_parameters ", data=bp_n_parameters)
            h5f.create_dataset("bp_n_measurements ", data=bp_n_measurements)
            h5f.create_dataset(
                "bp_n_rejected_measurements ", data=bp_n_rejected_measurements
            )
            h5f.create_dataset("bp_standard_deviation ", data=bp_standard_deviation)
            h5f.create_dataset("bp_chi_squared ", data=bp_chi_squared)
            if save_correlation_matrix:
                h5f.create_dataset(
                    "bp_coefficient_correlations ", data=bp_coefficient_correlations
                )
            h5f.create_dataset("bp_coefficient_errors", data=bp_coefficient_errors)
            h5f.create_dataset("bp_coefficients", data=bp_coefficients)
            h5f.create_dataset("bp_n_relevant_bases", data=bp_n_relevant_bases)
            h5f.create_dataset("bp_relative_shrinking", data=bp_relative_shrinking)

            h5f.create_dataset("rp_basis_function_id ", data=rp_basis_function_id)
            h5f.create_dataset("rp_degrees_of_freedom ", data=rp_degrees_of_freedom)
            h5f.create_dataset("rp_n_parameters ", data=rp_n_parameters)
            h5f.create_dataset("rp_n_measurements ", data=rp_n_measurements)
            h5f.create_dataset(
                "rp_n_rejected_measurements ", data=rp_n_rejected_measurements
            )
            h5f.create_dataset("rp_standard_deviation ", data=rp_standard_deviation)
            h5f.create_dataset("rp_chi_squared ", data=rp_chi_squared)
            if save_correlation_matrix:
                h5f.create_dataset(
                    "rp_coefficient_correlations ", data=rp_coefficient_correlations
                )
            h5f.create_dataset("rp_coefficient_errors", data=rp_coefficient_errors)
            h5f.create_dataset("rp_coefficients", data=rp_coefficients)
            h5f.create_dataset("rp_n_relevant_bases", data=rp_n_relevant_bases)
            h5f.create_dataset("rp_relative_shrinking", data=rp_relative_shrinking)
            h5f.close()


def compile_rvs_h5():
    for i_path in tqdm.tqdm(
        list(astro_data_path.joinpath("Gaia/gdr3/Spectroscopy/rvs_mean_spectrum/").glob(
            "*.csv.gz"
        )),
        desc="RVS spec",
    ):
        file_path_f = ascii.read(i_path)
        flux = np.vstack(file_path_f["flux"])
        flux_error = np.vstack(file_path_f["flux_error"])

        file_names_wo_ext = i_path.name[:-7]
        with h5py.File(f"./rvs_mean_spectrum/{file_names_wo_ext}.h5", "w") as f:
            f.create_dataset("source_id", data=file_path_f["source_id"].data)
            f.create_dataset("solution_id", data=file_path_f["solution_id"].data)
            f.create_dataset("ra", data=file_path_f["ra"].data)
            f.create_dataset("dec", data=file_path_f["dec"].data)
            f.create_dataset("flux", data=flux)
            f.create_dataset("flux_error", data=flux_error)


def compile_rvs_h5():
    for i_path in tqdm.tqdm(
        list(astro_data_path.joinpath("Gaia", "gdr3", "Spectroscopy", "rvs_mean_spectrum").glob(
            "*.csv.gz"
        )),
        desc="RVS spec",
    ):
        file_path_f = ascii.read(i_path)
        flux = np.vstack(file_path_f["flux"])
        flux_error = np.vstack(file_path_f["flux_error"])

        file_names_wo_ext = i_path.name[:-7]
        with h5py.File(f"./rvs_mean_spectrum/{file_names_wo_ext}.h5", "w") as f:
            f.create_dataset("source_id", data=file_path_f["source_id"].data)
            f.create_dataset("solution_id", data=file_path_f["solution_id"].data)
            f.create_dataset("ra", data=file_path_f["ra"].data)
            f.create_dataset("dec", data=file_path_f["dec"].data)
            f.create_dataset("flux", data=flux)
            f.create_dataset("flux_error", data=flux_error)


def comile_xp_mean_spec_h5():
    for i_path in tqdm.tqdm(
        list(astro_data_path.joinpath(
            "Gaia", "gdr3", "Spectroscopy", "xp_sampled_mean_spectrum"
        ).glob("*.csv.gz")),
        desc="XP specs",
    ):
        file_path_f = ascii.read(i_path)
        flux = np.vstack(file_path_f["flux"])
        flux_error = np.vstack(file_path_f["flux_error"])

        file_names_wo_ext = i_path.name[:-7]
        with h5py.File(f"./xp_sampled_mean_spectrum/{file_names_wo_ext}.h5", "w") as f:
            f.create_dataset("source_id", data=file_path_f["source_id"].data)
            f.create_dataset("solution_id", data=file_path_f["solution_id"].data)
            f.create_dataset("ra", data=file_path_f["ra"].data)
            f.create_dataset("dec", data=file_path_f["dec"].data)
            f.create_dataset("flux", data=flux)
            f.create_dataset("flux_error", data=flux_error)


def compile_gaia_sql_db(do_gaia_source_table=True, do_gaia_astrophysical_table=True, indexing=True):
    """
    This function compile Gaia SQL database
    """
    # The whole script takes about ~24 hours to complete
    Path(gaia_sql_db_path).touch()
    conn = sqlite3.connect(gaia_sql_db_path)
    c = conn.cursor()

    # we have added "grvs_mag" to the table on top of gaia_source_lite on Gaia Archive
    # will take ~11 hours to run
    if do_gaia_source_table:
        # =================== setup Gaia schema and first two tables ===================
        # this section will take ~30 minutes to run
        for schema in [
            "gaia_source_lite_schema",
            "allwise_best_neighbour_schema",
            "tmasspscxsc_best_neighbour_schema",
        ]:
            schema_filename = os.path.join(
                os.path.dirname(__file__), "sql_schema", f"{schema}"
            )
            with open(schema_filename) as f:
                lines = f.read().replace("\n", "")
            c.execute(lines)

        for name, table_name in zip(
            [_GAIA_DR3_ALLWISE_NEIGHBOUR_PARENT, _GAIA_DR3_2MASS_NEIGHBOUR_PARENT],
            ["allwise_best_neighbour", "tmasspscxsc_best_neighbour"],
        ):
            for p in tqdm.tqdm(list(name.glob("*.csv.gz"))):
                # load the data into a Pandas DataFrame
                data = pd.read_csv(p, header=0, sep=",")
                # write the data to a sqlite table
                data.to_sql(f"{table_name}", conn, if_exists="append", index=False)

        # =================== populate gaia_source lite table ===================
        # use "Int32" type for int columns with NaN
        for p in tqdm.tqdm(list(_GAIA_DR3_GAIASOURCE_PARENT.glob("*.csv.gz"))):
            dtypes = {
                "source_id": np.int64,
                "random_index": np.int64,
                "ra": np.float64,
                "ra_error": np.float64,
                "dec": np.float64,
                "dec_error": np.float64,
                "parallax": np.float64,
                "parallax_error": np.float64,
                "parallax_over_error": np.float32,
                "pmra": np.float64,
                "pmra_error": np.float64,
                "pmdec": np.float64,
                "pmdec_error": np.float64,
                "ra_dec_corr": np.float32,
                "ra_parallax_corr": np.float32,
                "ra_pmra_corr": np.float32,
                "ra_pmdec_corr": np.float32,
                "dec_parallax_corr": np.float32,
                "dec_pmra_corr": np.float32,
                "dec_pmdec_corr": np.float32,
                "parallax_pmra_corr": np.float32,
                "parallax_pmdec_corr": np.float32,
                "pmra_pmdec_corr": np.float32,
                "astrometric_params_solved": "Int32",
                "nu_eff_used_in_astrometry": np.float32,
                "pseudocolour": np.float64,
                "pseudocolour_error": np.float64,
                "astrometric_matched_transits": "Int32",
                "ipd_gof_harmonic_amplitude": np.float32,
                "ipd_frac_multi_peak": "Int32",
                "ipd_frac_odd_win": "Int32",
                "ruwe": np.float32,
                "phot_g_mean_flux": np.float64,
                "phot_g_mean_flux_over_error": np.float32,
                "phot_g_mean_mag": np.float32,
                "phot_bp_mean_flux": np.float64,
                "phot_bp_mean_flux_over_error": np.float32,
                "phot_bp_mean_mag": np.float32,
                "phot_rp_mean_flux": np.float64,
                "phot_rp_mean_flux_over_error": np.float32,
                "phot_rp_mean_mag": np.float32,
                "phot_bp_rp_excess_factor": np.float32,
                "bp_rp": np.float32,
                "radial_velocity": np.float64,
                "radial_velocity_error": np.float64,
                "rv_nb_transits": "Int32",
                "rv_expected_sig_to_noise": np.float32,
                "rv_renormalised_gof": np.float32,
                "rv_chisq_pvalue": np.float32,
                "rvs_spec_sig_to_noise": np.float32,
                "grvs_mag": np.float32,
                "l": np.float64,
                "b": np.float64,
                "has_xp_continuous": bool,
                "has_xp_sampled": bool,
                "has_rvs": bool,
            }
            data = pd.read_csv(
                p, header=1, sep=",", skiprows=999, usecols=dtypes.keys(), dtype=dtypes
            )
            # write the data to a sqlite table
            data.to_sql("gaia_source", conn, if_exists="append", index=False)

    if do_gaia_astrophysical_table:
        schema_filename = os.path.join(os.path.dirname(__file__), "sql_schema", "astrophysical_parameters_lite_schema")

        with open(schema_filename) as f:
            lines = f.read().replace("\n", "")
        c.execute(lines)
        # =================== populate gaia_source lite table ===================
        # we have added "grvs_mag" to the table on top of gaia_source_lite on Gaia Archive
        # will take ~11 hours to run

        for p in tqdm.tqdm(list(_GAIA_DR3_ASTROPHYS_PARENT.glob("*.csv.gz"))):
            dtypes = {
                "source_id": np.int64,
                "classprob_dsc_combmod_quasar": np.float32,
                "classprob_dsc_combmod_galaxy": np.float32,
                "classprob_dsc_combmod_star": np.float32,
                "classprob_dsc_combmod_whitedwarf": np.float32,
                "classprob_dsc_combmod_binarystar": np.float32,
                "classprob_dsc_specmod_quasar": np.float32, 
                "classprob_dsc_specmod_galaxy": np.float32,
                "classprob_dsc_specmod_star": np.float32,
                "classprob_dsc_specmod_whitedwarf": np.float32,
                "classprob_dsc_specmod_binarystar": np.float32,
                "classprob_dsc_allosmod_quasar": np.float32,
                "classprob_dsc_allosmod_galaxy": np.float32,
                "classprob_dsc_allosmod_star": np.float32,
                "teff_gspphot": np.float32,
                "teff_gspphot_lower": np.float32,
                "teff_gspphot_upper": np.float32,
                "logg_gspphot": np.float32,
                "logg_gspphot_lower": np.float32,
                "logg_gspphot_upper": np.float32,
                "mh_gspphot": np.float32,
                "mh_gspphot_lower": np.float32,
                "mh_gspphot_upper": np.float32,
                "distance_gspphot": np.float32,
                "distance_gspphot_lower": np.float32,
                "distance_gspphot_upper": np.float32,
                "azero_gspphot": np.float32,
                "azero_gspphot_lower": np.float32,
                "azero_gspphot_upper": np.float32,
                "ag_gspphot": np.float32,
                "ag_gspphot_lower": np.float32,
                "ag_gspphot_upper": np.float32,
                "abp_gspphot": np.float32,
                "abp_gspphot_lower": np.float32,
                "abp_gspphot_upper": np.float32,
                "arp_gspphot": np.float32,
                "arp_gspphot_lower": np.float32,
                "arp_gspphot_upper": np.float32,
                "ebpminrp_gspphot": np.float32,
                "ebpminrp_gspphot_lower": np.float32,
                "ebpminrp_gspphot_upper": np.float32,
                "mg_gspphot": np.float32,
                "mg_gspphot_lower": np.float32,
                "mg_gspphot_upper": np.float32,
                "radius_gspphot": np.float32,
                "radius_gspphot_lower": np.float32,
                "radius_gspphot_upper": np.float32,
                "logposterior_gspphot": np.float32,
                "mcmcaccept_gspphot": np.float32,
                "libname_gspphot": str,
                "teff_gspspec": np.float32,
                "teff_gspspec_lower": np.float32,
                "teff_gspspec_upper": np.float32,
                "logg_gspspec": np.float32,
                "logg_gspspec_lower": np.float32,
                "logg_gspspec_upper": np.float32,
                "mh_gspspec": np.float32,
                "mh_gspspec_lower": np.float32,
                "mh_gspspec_upper": np.float32,
                "alphafe_gspspec": np.float32,
                "alphafe_gspspec_lower": np.float32,
                "alphafe_gspspec_upper": np.float32,
                "flags_gspspec": str,
                "activityindex_espcs": np.float32,
                "activityindex_espcs_uncertainty": np.float32,
                "activityindex_espcs_input": str
            }
            data = pd.read_csv(
                p, header=1, sep=",", skiprows=1540, usecols=dtypes.keys(), dtype=dtypes
            )
            # write the data to a sqlite table
            data.to_sql("astrophysical_parameters", conn, if_exists="append", index=False)

    # =================== indexing ===================
    if indexing:
        print("=================== indexing ===================")
        print("Start doing allwise_best_neighbour_sourceid_designation indexing")
        c.execute(
            """CREATE INDEX allwise_best_neighbour_sourceid_designation ON allwise_best_neighbour (source_id, original_ext_source_id);"""
        )
        print("Start doing tmasspscxsc_best_neighbour_sourceid_designation indexing")
        c.execute(
            """CREATE INDEX tmasspscxsc_best_neighbour_sourceid_designation ON tmasspscxsc_best_neighbour (source_id, original_ext_source_id);"""
        )


def compile_tmass_sql_db(indexing=True):
    """
    This function compile 2MASS point source SQL database
    """
    Path(tmass_sql_db_path).touch()
    conn = sqlite3.connect(tmass_sql_db_path)
    c = conn.cursor()

    # =================== 2MASS ===================
    # this section will take 1 hour to run
    schema_filename = os.path.join(os.path.dirname(__file__), "sql_schema", "twomass_psc_lite_schema")

    with open(schema_filename) as f:
        lines = f.read().replace("\n", "")
    c.execute(lines)

    # only the first part, not all actually
    # https://irsa.ipac.caltech.edu/data/2MASS/docs/releases/allsky/doc/sec2_2a.html
    tmass_allcol = [
        "ra",
        "dec",
        "err_maj",
        "err_min",
        "err_ang",
        "designation",
        # Primary Photometric Information
        "j_m",
        "j_cmsig",
        "j_msigcom",
        "j_snr",
        "h_m",
        "h_cmsig",
        "h_msigcom",
        "h_snr",
        "k_m",
        "k_cmsig",
        "k_msigcom",
        "k_snr",
        # Primary Source Quality Information
        "ph_qual",
        "rd_flg",
        "bl_flg",
        "cc_flg",
        "ndet",
        "prox",
        "pxpa",
        "pxcntr",
        "gal_contam",
        "mp_flg",
        # Additional Positional and Identification Information
        "pts_key/cntr",
        "hemis",
        "date",
        "scan",
        "glon",
        "glat",
        "x_scan",
        "jdate",
        # Additional Photometric Information
        "j_psfchi",
        "h_psfchi",
        "k_psfchi"
        "j_m_stdap",
        "j_msig_stdap",
        "h_m_stdap",
        "h_msig_stdap",
        "k_m_stdap",
        "k_msig_stdap",
        # Additional Source Quality Information
        "dist_edge_ns",
        "dist_edge_ew",
        "dist_edge_flg",
        "dup_src",
        "use_src",
        # Optical Source Association Information
        "a",
        "dist_opt",
        "phi_opt",
        "b_m_opt",
        "vr_m_opt",
        "nopt_mchs",
        # Cross-Index Information
        "ext_key",
        "scan_key",
        "coadd_key",
        "coadd"
    ]

    for p in tqdm.tqdm(list(_2MASS_PARENT.glob("psc_*.gz"))):
        dtypes = {
            "ra": np.float64,
            "dec": np.float64,
            "designation": str,
            "j_m": np.float32,
            "j_cmsig": np.float32,
            "j_msigcom": np.float32,
            "j_snr": np.float32,
            "h_m": np.float32,
            "h_cmsig": np.float32,
            "h_msigcom": np.float32,
            "h_snr": np.float32,
            "k_m": np.float32,
            "k_cmsig": np.float32,
            "k_msigcom": np.float32,
            "k_snr": np.float32,
            "ph_qual": str,
            "rd_flg": str,
            "bl_flg": str,
            "cc_flg": str,
            "ndet": str,
            "prox": np.float32,
        }
        data = pd.read_csv(
            p,
            header=None,
            sep="|",
            usecols=[tmass_allcol.index(i) for i in dtypes.keys()],
            names=dtypes.keys(),
            # turn null to proper NaN
            na_values=["\\N"],
            # dont allow white space in names since gaia best neightbour do not have white space
            converters={"designation": str.strip},
        )
        data = data.replace(r"\N", np.nan)
        data.astype(dtypes)

        # write the data to a sqlite table
        data.to_sql(f"twomass_psc", conn, if_exists="append", index=False)

    # =================== indexing ===================
    if indexing:
        # 9m46s
        print("Doing Indexing")
        c.execute(
            """CREATE INDEX twomass_psc_designation_mags ON twomass_psc (designation, j_m, h_m, k_m);"""
        )


def compile_allwise_sql_db(indexing=True):
    """
    This function compile allwise SQL database
    """
    Path(allwise_sql_db_path).touch()
    conn = sqlite3.connect(allwise_sql_db_path)
    c = conn.cursor()

    # this section will take ~16 hours to run
    schema_filename = os.path.join(os.path.dirname(__file__), "sql_schema", "allwise_lite_schema")

    with open(schema_filename) as f:
        lines = f.read().replace("\n", "")
    c.execute(lines)

    # only the first part, not all actually
    # https://wise2.ipac.caltech.edu/docs/release/allwise/expsup/sec2_1a.html
    allwise_allcol = [
        # Basic Position and Identification Information
        "designation",
        "ra",
        "dec",
        "sigra",
        "sigdec",
        "sigradec",
        "glon",
        "glat",
        "elon",
        "elat",
        "wx",
        "wy",
        "cntr",
        "source_id",
        "coadd_id",
        "src",
        # Primary Photometric Information
        "w1mpro",
        "w1sigmpro",
        "w1snr",
        "w1rchi2",
        "w2mpro",
        "w2sigmpro",
        "w2snr",
        "w2rchi2",
        "w3mpro",
        "w3sigmpro",
        "w3snr",
        "w3rchi2",
        "w4mpro",
        "w4sigmpro",
        "w4snr",
        "w4rchi2",
        "rchi2",
        "nb",
        "na",
        "w1sat",
        "w2sat",
        "w3sat",
        "w4sat",
        "satnum",
        # Motion Fit Parameters
        "ra_pm",
        "dec_pm",
        "sigra_pm",
        "sigdec_pm",
        "sigradec_pm",
        "pmra",
        "sigpmra",
        "pmdec",
        "sigpmdec",
        "w1rchi2_pm",
        "w2rchi2_pm",
        "w3rchi2_pm",
        "w4rchi2_pm",
        "rchi2_pm",
        "pmcode",
        # Measurement Quality and Source Reliability Information
        "cc_flags",
        "rel",
        "ext_flg",
        "var_flg",
        "ph_qual",
        "det_bit",
        "moon_lev",
        "w1nm",
        "w1m",
        "w2nm",
        "w2m",    
        "w3nm",
        "w3m",    
        "w4nm",
        "w4m",
        "w1cov",
        "w2cov",
        "w3cov",
        "w4cov",
        "w1cc_map",
        "w1cc_map_str",
        "w2cc_map",
        "w2cc_map_str",
        "w3cc_map",
        "w3cc_map_str",
        "w4cc_map",
        "w4cc_map_str",
        "use_src",
        "best_use_src", 
        "ngrp",
        # Additional Photometric Information
        "w1flux",
        "w1sigflux",
        "w1sky",
        "w1sigsk",
        "w1conf",
        "w2flux",
        "w2sigflux",
        "w2sky",
        "w2sigsk",
        "w2conf",
        "w3flux",
        "w3sigflux",
        "w3sky",
        "w3sigsk",
        "w3conf",
        "w4flux",
        "w4sigflux",
        "w4sky",
        "w4sigsk",
        "w4conf",
        "w1mag",
        "w1sigm",
        "w1flg",
        "w1mcor",
        "w2mag",
        "w2sigm",
        "w2flg",
        "w2mcor",
        "w3mag",
        "w3sigm",
        "w3flg",
        "w3mcor",
        "w4mag",
        "w4sigm",
        "w4flg",
        "w4mcor",
        "w1mag_1",
        "w1sigm_1",
        "w1flg_1",
        "w2mag_1",
        "w2sigm_1",
        "w2flg_1",
        "w3mag_1",
        "w3sigm_1",
        "w3flg_1",
        "w4mag_1",
        "w4sigm_1",
        "w4flg_1",
        "w1mag_2",
        "w1sigm_2",
        "w1flg_2",
        "w2mag_2",
        "w2sigm_2",
        "w2flg_2",
        "w3mag_2",
        "w3sigm_2",
        "w3flg_2",
        "w4mag_2",
        "w4sigm_2",
        "w4flg_2",
        "w1mag_3",
        "w1sigm_3",
        "w1flg_3",
        "w2mag_3",
        "w2sigm_3",
        "w2flg_3",
        "w3mag_3",
        "w3sigm_3",
        "w3flg_3",
        "w4mag_3",
        "w4sigm_3",
        "w4flg_3",
        "w1mag_4",
        "w1sigm_4",
        "w1flg_4",
        "w2mag_4",
        "w2sigm_4",
        "w2flg_4",
        "w3mag_4",
        "w3sigm_4",
        "w3flg_4",
        "w4mag_4",
        "w4sigm_4",
        "w4flg_4",
        "w1mag_5",
        "w1sigm_5",
        "w1flg_5",
        "w2mag_5",
        "w2sigm_5",
        "w2flg_5",
        "w3mag_5",
        "w3sigm_5",
        "w3flg_5",
        "w4mag_5",
        "w4sigm_5",
        "w4flg_5",
        "w1mag_6",
        "w1sigm_6",
        "w1flg_6",
        "w2mag_6",
        "w2sigm_6",
        "w2flg_6",
        "w3mag_6",
        "w3sigm_6",
        "w3flg_6",
        "w4mag_6",
        "w4sigm_6",
        "w4flg_6",
        "w1mag_7",
        "w1sigm_7",
        "w1flg_7",
        "w2mag_7",
        "w2sigm_7",
        "w2flg_7",
        "w3mag_7",
        "w3sigm_7",
        "w3flg_7",
        "w4mag_7",
        "w4sigm_7",
        "w4flg_7",
        "w1mag_",
        "w1sigm_8",
        "w1flg_8",
        "w2mag_8",
        "w2sigm_8",
        "w2flg_8",
        "w3mag_8",
        "w3sigm_8",
        "w3flg_8",
        "w4mag_8",
        "w4sigm_8",
        "w4flg_8",
        "w1magp",
        "w1sigp1",
        "w1sigp2",
        "w1k",
        "w1ndf",
        "w1mlq",
        "w1mjdmin",
        "w1mjdmax",
        "w1mjdmean",
        "w2magp",
        "w2sigp1",
        "w2sigp2",
        "w2k",
        "w2ndf",
        "w2mlq",
        "w2mjdmin",
        "w2mjdmax",
        "w2mjdmean",
        "w3magp",
        "w3sigp1",
        "w3sigp2",
        "w3k",
        "w3ndf",
        "w3mlq",
        "w3mjdmin",
        "w3mjdmax",
        "w3mjdmean",
        "w4magp",
        "w4sigp1",
        "w4sigp2",
        "w4k",
        "w4ndf",
        "w4mlq",
        "w4mjdmin",
        "w4mjdmax",
        "w4mjdmean",
        "rho12",
        "rho23",
        "rho34",
        "q12",
        "q23",
        "q34",
        "xscprox",
        "w1rsemi",
        "w1ba",
        "w1pa",
        "w1gmag",
        "w1gerr",
        "w1gflg",
        "w2rsemi",
        "w2ba",
        "w2pa",
        "w2gmag",
        "w2gerr",
        "w2gflg",
        "w3rsemi",
        "w3ba",
        "w3pa",
        "w3gmag",
        "w3gerr",
        "w3gflg",
        "w4rsemi",
        "w4ba",
        "w4pa",
        "w4gmag",
        "w4gerr",
        "w4gflg",
        "tmass_key",
        "r_2mass",
        "pa_2mass",
        "n_2mass",
        "j_m_2mass",
        "j_msig_2mass",
        "h_m_2mass",
        "h_msig_2mass",
        "k_m_2mass",
        "k_msig_2mass",
        "x",
        "y",
        "z",
        "spt_ind",
        "htm20",
    ]

    for p in tqdm.tqdm(list(_ALLWISE_PARENT.glob("wise-allwise-cat-*.bz2"))):
        dtypes = {
            "designation": str,  # designation does not seems to be unique, dont use it as primary key
            "ra": np.float64,
            "dec": np.float64,
            "sigra": np.float32,
            "sigdec": np.float32,
            "sigradec": np.float32,
            "w1mpro": np.float32,
            "w1sigmpro": np.float32,
            "w1snr": np.float32,
            "w2mpro": np.float32,
            "w2sigmpro": np.float32,
            "w2snr": np.float32,
            "w3mpro": np.float32,
            "w3sigmpro": np.float32,
            "w3snr": np.float32,
            "w4mpro": np.float32,
            "w4sigmpro": np.float32,
            "w4snr": np.float32,
            "nb": "Int32",
            "na": "Int32",
            "cc_flags": str,
            "ext_flg": "Int32",
            "var_flg": str,
            "ph_qual": str,
            "w1mjdmean": np.float64,
            "w2mjdmean": np.float64,
            "w3mjdmean": np.float64,
            "w4mjdmean": np.float64,
            "w1gmag": np.float32,
            "w1gerr": np.float32,
            "w2gmag": np.float32,
            "w2gerr": np.float32,
            "w3gmag": np.float32,
            "w3gerr": np.float32,
            "w4gmag": np.float32,
            "w4gerr": np.float32,
        }
        data = pd.read_csv(
            p,
            header=None,
            sep="|",
            usecols=[allwise_allcol.index(i) for i in dtypes.keys()],
            names=dtypes.keys(),
        )
        data.astype(dtypes)
        
        # write the data to a sqlite table
        data.to_sql(f"allwise", conn, if_exists="append", index=False)

    # =================== indexing ===================
    if indexing:
        # 22m56s
        print("Doing Indexing")
        c.execute(
            """CREATE INDEX allwise_designation_mags ON allwise (designation, w1mpro, w2mpro, w3mpro, w4mpro, w1snr, w2snr, w3snr, w4snr, ph_qual);"""
        )



import os
import pathlib
from setuptools import Extension, setup, find_packages

with open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.rst"),
    encoding="utf-8",
) as f:
    long_description = f.read()

astroqlite_c_src = pathlib.Path("./mygaiadb/ext_c/astroqlite.c")

astroqlite_c = Extension(
    "astroqlite_c",
    sources=[astroqlite_c_src.as_posix()],
    include_dirs=[pathlib.Path("./mygaiadb/ext_c/").as_posix()],
)

ext_modules = [astroqlite_c]

setup(
    name="MyGaiaDB",
    version="0.2",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Database :: Front-Ends",
    ],
    packages=find_packages(),
    package_data={"mygaiadb/sql_schema": ["*"]},
    include_package_data=True,
    python_requires=f">=3.8",
    install_requires=[
        "numpy",
        "astropy",
        "h5py",
        "pandas",
        "tqdm",
    ],
    ext_modules=ext_modules,
    url="https://github.com/henrysky/MyGaiaDB",
    project_urls={
        "Bug Tracker": "https://github.com/henrysky/MyGaiaDB/issues",
        "Documentation": "https://github.com/henrysky/MyGaiaDB",
        "Source Code": "https://github.com/henrysky/MyGaiaDB",
    },
    license="MIT",
    author="Henry Leung",
    author_email="henrysky.leung@utoronto.ca",
    description="Setup local serverless ESA Gaia / 2MASS / ALLWISE databases and run query locally with python",
    long_description=long_description,
)



import sqlite3
import numpy as np
from mygaiadb.query import LocalGaiaSQL


def test_c_exts():
    """
    Test SQL C extensions are working as expected compared to numpy and known values
    """
    c = sqlite3.connect(":memory:")
    LocalGaiaSQL._load_sqlite3_ext(c)
    # test geometrical functions
    np.testing.assert_approx_equal(c.execute("""SELECT COS(1.)""").fetchall()[0][0], np.cos(1.), 14)
    np.testing.assert_approx_equal(c.execute("""SELECT SIN(1.)""").fetchall()[0][0], np.sin(1.), 14)
    np.testing.assert_approx_equal(c.execute("""SELECT TAN(1.)""").fetchall()[0][0], np.tan(1.), 14)

    # make sure random is actually random
    rand_1 = c.execute("""SELECT RAND()""").fetchall()
    rand_2 = c.execute("""SELECT RAND()""").fetchall()
    assert rand_1 != rand_2

    # merak and dubhe for all amateur stargazer
    np.testing.assert_approx_equal(c.execute("""SELECT DISTANCE(165.458, 56.3825, 165.933, 61.7511)""").fetchall()[0][0], 5.37411, 4)
    np.testing.assert_approx_equal(c.execute("""SELECT DISTANCE(165.933, 61.7511, 165.458, 56.3825)""").fetchall()[0][0], 5.37411, 4)
    # known bad G.source_id = 4472832130942575872
    np.testing.assert_approx_equal(c.execute("""SELECT DISTANCE(179., 10., 269.448503, 4.73942)""").fetchall()[0][0], 89.6181177, 4)

    # examples on https://www.cosmos.esa.int/web/gaia-users/archive/writing-queries#adql_syntax_2
    np.testing.assert_approx_equal(c.execute("""SELECT CBRT(27.)""").fetchall()[0][0], 3., 15)

    # examples on https://www.cosmos.esa.int/web/gaia-users/archive/writing-queries#adql_syntax_1
    np.testing.assert_approx_equal(c.execute("""SELECT GAIA_HEALPIX_INDEX(4, 2060294888487267584)""").fetchall()[0][0], 914, 15)
    np.testing.assert_approx_equal(c.execute("""SELECT GAIA_HEALPIX_INDEX(8, 539853444666465792)""").fetchall()[0][0], 61374, 15)
    np.testing.assert_approx_equal(c.execute("""SELECT GAIA_HEALPIX_INDEX(8, 1678668421246394624)""").fetchall()[0][0], 190842, 15)
    np.testing.assert_approx_equal(c.execute("""SELECT SIGN(-0.001)""").fetchall()[0][0], -1, 15)
